# Package Info
NAME = 'JAIS'
VERSION = '0.0.1.13'
DESCRIPTION = "Just Artificial Intelligence Snippets (JAIS)"

# Basic settings
from pathlib import Path
ROOT_DIR = Path(__file__).parent  # Package path
PROJ_DIR = ROOT_DIR.parent  # Project path